// ///////////////////////////////////////////////////////////////////////////
//
//  DS1410D.c                                 { Ver 2.54  03/14/97 }
//
//    Windows NT driver for the DS1410E parallel port resident adapter 
//
//  Revisions
//      1.00 -> 1.01   Bit wiggling stuff converted to DS1410D operation
//                   only.
// 
//      1.01 -> 2.00   Added code for support of the DS1410E with EPP
//                   passthru mode. DS1410D adapters are still supported. 
//
//      2.00 -> 2.50   Added block I/O functionality to minimize context
//                   switching for CiB communication. Also added a fast
//                   access which should only be used when communicating
//                   in overdrive !!!
// 
//      2.50 -> 2.51   Added check overdrive function to test overdrive
//                   without affecting the state of the overdrive FF in  
//                   DS1481. Also added a fast byte routine witch bangs
//                   8 bits without restoring the old IRQL. This should
//                   only be used in overdrive mode !!!
//
//      2.51 -> 2.52   Added dynamic unloading capability.
//
//      2.52 -> 2.53   Left 1 of the 6 high order data bits low after
//                  a docomm time slot to avoid problems with anything
//                  that grounds 14. This was the case up until ver 2.5
//                  at which time the lines were all driven high to help
//                  power the CiB.
//
//      2.53 -> 2.54   Change all bit-wiggling code to ensure that we can
//                  speak to a DS1410E even if the printer is grounding
//                  pin 14.
//
//
//  Copyright (C) 1994, 96, 97, 98, 99 Dallas Semiconductor Corporation. 
//  All rights Reserved. Printed in U.S.A.
//  This software is protected by copyright laws of
//  the United States and of foreign countries.
//  This material may also be protected by patent laws of the United States 
//  and of foreign countries.
//  This software is furnished under a license agreement and/or a
//  nondisclosure agreement and may only be used or copied in accordance
//  with the terms of those agreements.
//  The mere transfer of this software does not imply any licenses
//  of trade secrets, proprietary technology, copyrights, patents,
//  trademarks, maskwork rights, or any other form of intellectual
//  property whatsoever. Dallas Semiconductor retains all ownership rights.
//
/////////////////////////////////////////////////////////////////////////////

#include <stddef.h>
#include "ntddk.h"

#include "ds1410d.h"

#define NAND_MASK 0xAB
#define CONTROL_REG_MASK 0x1F


/////////////////////////////////////////////////////////////////////////////
//
//    Create Device Object, Create symbolic link, then register driver 
// entry points.
//
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,  
                     IN PUNICODE_STRING RegistryPath)
{
   NTSTATUS              status;           // Return value of this function
   STRING                AnsiUserName;     // Ansi user name
   STRING                AnsiObjectName;   // Ansi object name
   UNICODE_STRING        UserName;         // Unicode user name
   UNICODE_STRING        ObjectName;       // Unicode object name
   PDEVICE_OBJECT        deviceObject;     // Pointer to device object
   DOW_CONFIG_INFO       ConfigInfo;       // Pointer to config structure
   PDOW_DEVICE_EXTENSION extension;        // Pointer to extension 
   BOOLEAN               ConflictDetected; 
   CONFIGURATION_TYPE    ControllerType;   // Controller type to search for
   INTERFACE_TYPE        interfaceType;    // ISA, EISA, MICROCHANNEL, ...
   ULONG                 PortsFound;       // Number of parallel ports found
   ULONG                 ControllerNumber; // Specifies 0 based controller #

   PortsFound = 0;                         
   // Controller type to search for
   ControllerType = ParallelController;       

   RtlInitString(&AnsiUserName, "\\DosDevices\\LPT9"); 
   RtlInitString(&AnsiObjectName, "\\Device\\DOW0");

   // Convert user name to unicode string
   status = RtlAnsiStringToUnicodeString(&UserName, 
                                         &AnsiUserName,
                                         TRUE);
   // Convert object name to unicode string 
   if (status == STATUS_SUCCESS)      
      status = RtlAnsiStringToUnicodeString(&ObjectName, 
                                            &AnsiObjectName,
                                            TRUE);

   // If successful create device object
   if (status == STATUS_SUCCESS)       
   {
      status = IoCreateDevice(DriverObject, 
                              sizeof(DOW_DEVICE_EXTENSION),
                              &ObjectName,               
                              FILE_DEVICE_PARALLEL_PORT,
                              0,
                              TRUE,       // Only 1 thread at a time please 
                              &deviceObject);              

      if (status == STATUS_SUCCESS)           
      {
         // Get extension pointer
         extension = deviceObject->DeviceExtension;   
         deviceObject->Flags |= DO_BUFFERED_IO; 

         PortsFound = 0;                    

         // Start at bus type 0
         interfaceType = 0;                             

         //
         // Loop over all buses (unless we find to many ports)
         //
         do              
         {
            // Assume no ports on this bus
            ConfigInfo.FoundOneMan = FALSE;  

            // Start at 0th controller
            ControllerNumber = 0;                
              
            do
            {
               // Assume no port number ControllerNumber
               ConfigInfo.FoundOneMan = FALSE; 

               // Get port info
               IoQueryDeviceDescription(&interfaceType,    
                                        NULL,
                                        &ControllerType,
                                        &ControllerNumber,
                                        NULL,
                                        NULL,
                                        DOWConfigCallBack,
                                        (PVOID) &ConfigInfo);
         
               if (ConfigInfo.FoundOneMan)   
               {
                  // Memory map this thing
                  extension->Controller[PortsFound] = 
                     DOWGetMmAddress(ConfigInfo.InterfaceType,
                                     ConfigInfo.BusNumber,
                                     ConfigInfo.ParControllerAddress,
                                     ConfigInfo.SpanOfController,
                                     (BOOLEAN) ConfigInfo.AddressSpace,
                                     &extension->UnMapRegisters);
 
                  // Store controller span in device extension
                  extension->SpanOfController = ConfigInfo.SpanOfController;

                  ++PortsFound;
                  ++ControllerNumber;
               }
            }
            while ((PortsFound < MAX_PAR_NUM) && ConfigInfo.FoundOneMan);
         }
         while ((PortsFound < MAX_PAR_NUM) && 
                (interfaceType++ < MaximumInterfaceType));
      }

      if (PortsFound > 0)   
      {
         // Save # of ports found
         extension->ParNumFound = (UCHAR) PortsFound; 

         // Was mapping of at least 1 port good ?
         if (!extension->Controller[0])  
            status = STATUS_NO_SUCH_DEVICE;

         // Create symbolic link
         if (status == STATUS_SUCCESS)                 
         {
            extension->pn = 0;                        

            status = IoCreateUnprotectedSymbolicLink(&UserName, &ObjectName);

            // Report resource usage to registry
            IoReportResourceUsage(NULL,   
                                  DriverObject,
                                  NULL,
                                  0,
                                  deviceObject,
                                  NULL,
                                  0,
                                  FALSE,
                                  &ConflictDetected);

            // Not much chance of this since were not claiming an interrupt
            if (ConflictDetected) 
               status = STATUS_INSUFFICIENT_RESOURCES; 
         }
      }
      else
         status = STATUS_NO_SUCH_DEVICE; // Didn't find any parallel ports

      // Release memory held by UserName
      RtlFreeUnicodeString(&UserName);  
      // Release memory held by ObjectName
      RtlFreeUnicodeString(&ObjectName); 
   }
   else
      RtlFreeUnicodeString(&UserName); 

   // Register driver entry points
   if (status == STATUS_SUCCESS)               
   {
      DriverObject->MajorFunction[IRP_MJ_WRITE]          = DS1410XDispatch;
      DriverObject->MajorFunction[IRP_MJ_READ]           = DS1410XDispatch;
      DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DS1410XDispatch;
      DriverObject->MajorFunction[IRP_MJ_CREATE]         = DS1410XDispatch;
      DriverObject->MajorFunction[IRP_MJ_CLOSE]          = DS1410XDispatch;
      DriverObject->DriverUnload                         = DOWDriverUnload;
   }

   // return result to the I/O manager
   return status;                          
}

/////////////////////////////////////////////////////////////////////////////
//
//    Configuration call back routine. We get to learn about the hardware
// on which wer'e running.
//
NTSTATUS DOWConfigCallBack(IN PVOID Context,
                           IN PUNICODE_STRING PathName,
                           IN INTERFACE_TYPE BusType,
                           IN ULONG BusNumber,
                           IN PKEY_VALUE_FULL_INFORMATION *BusInfo,
                           IN CONFIGURATION_TYPE ControllerType,
                           IN ULONG ControllerNumber,
                           IN PKEY_VALUE_FULL_INFORMATION *ControllerInfo,
                           IN CONFIGURATION_TYPE PeripheralType,
                           IN ULONG PeripheralNumber,
                           IN PKEY_VALUE_FULL_INFORMATION *PeripheralInfo)
{
   PDOW_CONFIG_INFO config = Context;
   ULONG i;
   PCM_FULL_RESOURCE_DESCRIPTOR controllerData;

   controllerData =
       (PCM_FULL_RESOURCE_DESCRIPTOR)
       (((PUCHAR)ControllerInfo[IoQueryDeviceConfigurationData]) +
        ControllerInfo[IoQueryDeviceConfigurationData]->DataOffset);

   config->InterfaceType = BusType;
   config->BusNumber = BusNumber;

   for (i = 0; i < controllerData->PartialResourceList.Count; i++) 
   {
      PCM_PARTIAL_RESOURCE_DESCRIPTOR partial =
          &controllerData->PartialResourceList.PartialDescriptors[i];

      switch (partial->Type)  
      {
          case CmResourceTypePort: 
              // Looks like we got one
              config->FoundOneMan = TRUE;             
              // Data status control
              config->SpanOfController = 3;           
              config->ParControllerAddress = partial->u.Port.Start; 
              // Memory or port I/O
              config->AddressSpace = partial->Flags;     
          break;

          default:              
          break;
      }
   }

   return STATUS_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
//
//    Get a memory mapped address for the HAL to use for our bit-banging
// needs.
//
PVOID DOWGetMmAddress(IN INTERFACE_TYPE BusType,
                      IN ULONG BusNumber,
                      PHYSICAL_ADDRESS IoAddress,
                      ULONG NumberOfBytes,
                      ULONG AddressSpace,
                      PBOOLEAN UMap)
{
   PHYSICAL_ADDRESS TranslatedAddress;
   PVOID address;

   if (!HalTranslateBusAddress(BusType,
                               BusNumber,
                               IoAddress,
                               &AddressSpace,
                               &TranslatedAddress))
   {
      AddressSpace = 1;
   }

   // If address not in virtual space, put it there
   if (!AddressSpace)         
   {
      address = MmMapIoSpace(TranslatedAddress, NumberOfBytes, FALSE);
      *UMap   = (address ? TRUE : FALSE);
   }
   else 
   {
      address = (PVOID) TranslatedAddress.LowPart;
      *UMap   = FALSE;
   }

   return address;
}

/////////////////////////////////////////////////////////////////////////////
//
//    Undo all the stuff done at driver initialization. 
//
VOID DOWDriverUnload(IN PDRIVER_OBJECT DOWDriverObject)
{
   STRING         AnsiObjectName;   // Ansi string object name
   UNICODE_STRING ObjectName;       // Unicode string object name
   PDEVICE_OBJECT CurrentDevice;
   UCHAR          i;

   RtlInitString(&AnsiObjectName, "\\DosDevices\\LPT9");

   // Convert object name to unicode string 
   RtlAnsiStringToUnicodeString(&ObjectName, 
                                &AnsiObjectName,
                                TRUE);

   // Whack the symbolic link
   IoDeleteSymbolicLink(&ObjectName);

   i = 0;
   while (CurrentDevice = DOWDriverObject->DeviceObject)
   {
      DOWCleanUpDevice(CurrentDevice->DeviceExtension, i++);
      // Undo IoCreateDevice 
      IoDeleteDevice(CurrentDevice);
   }
}

/////////////////////////////////////////////////////////////////////////////
//
//    Unmap translated bus addresses
//
VOID DOWCleanUpDevice(IN PDEVICE_OBJECT DOWDeviceObject, IN UCHAR i)
{
   PDOW_DEVICE_EXTENSION extension = DOWDeviceObject->DeviceExtension;

   if (extension && extension->UnMapRegisters)
   {
   /*
      MmUnmapIoSpace(extension->Controller[i],
                     3); 
                     //extension->SpanOfController); 
   */
   }
}
 
/////////////////////////////////////////////////////////////////////////////
//
//    Commands processed here.
//
NTSTATUS DS1410XDispatch(IN PDEVICE_OBJECT ODevObject, IN PIRP OIrp)
{
   UCHAR              APtr, PN;
   UCHAR              tbit, tbyte, rbyte;
   UCHAR              i;
   ULONG              j;
   NTSTATUS           status;
   PIO_STACK_LOCATION OIrpSp;

   // Get pointer to system buffer for I/O
   PUCHAR irpBuffer = OIrp->AssociatedIrp.SystemBuffer;
   // Get extension so we can have our globals
   PDOW_DEVICE_EXTENSION extension = ODevObject->DeviceExtension;

   OIrpSp = IoGetCurrentIrpStackLocation(OIrp);              

   OIrp->IoStatus.Information = 0L;               
   status = STATUS_SUCCESS;                 

   switch (OIrpSp->MajorFunction)
   {
      // Called by CloseHandle
      case IRP_MJ_CLOSE:  
      break;

      // Called by CreateFile
      case IRP_MJ_CREATE: 
         OIrp->IoStatus.Status= status; 
      break;

      // Called by ReadFile
      case IRP_MJ_READ:   
         if ((OIrpSp->Parameters.Read.ByteOffset.HighPart != 0) ||
             (OIrpSp->Parameters.Read.ByteOffset.LowPart != 0))
         {
            status = STATUS_INVALID_PARAMETER;
         }
         else
         {
            if (OIrpSp->Parameters.Read.Length == extension->CRLen)
            {
               // Send back result of last Write
               for (j = 0; j < extension->CRLen; j++)
                  irpBuffer[j] = extension->CommResult[j];
      
               OIrpSp->Parameters.Read.Length = 0; 
               OIrp->IoStatus.Information = extension->CRLen;     

               OIrp->IoStatus.Status = STATUS_SUCCESS; 
            }
            else
               status = STATUS_INVALID_PARAMETER;
         }
      break;

      // Called by WriteFile
      case IRP_MJ_WRITE:   
         if ((OIrpSp->Parameters.Write.ByteOffset.HighPart != 0) ||
             (OIrpSp->Parameters.Write.ByteOffset.LowPart != 0))
         {
            status = STATUS_INVALID_PARAMETER;
         }
         else
         {
            if (OIrpSp->Parameters.Write.Length >= 2)
            {
               // Default return length 
               extension->CRLen = 1;

               switch (irpBuffer[0])
               {
                  case DO_RESET:                     
                     // Get port number
                     APtr = irpBuffer[1];       

                     // Is this a good # 
                     if (extension->ParNumFound >= APtr)   
                     { 
                        // Set new (0 based) port number 
                        extension->pn = APtr - 1;       

                        // Do a reset
                        extension->CommResult[0] =                  
                           docomm(DOW_RESET, 
                                  extension->Controller[APtr - 1]) ^ 1;

                        // Indicate last port good 
                        extension->GoodPort = TRUE; 
                     }
                     else
                     { 
                        // Indicate last port bad
                        extension->GoodPort = FALSE; 
                        // Fail if invalid port #
                        extension->CommResult[0] = 0;   
                     }
                  break;

                  case DO_BIT:                          
                     // Are we on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         
                        // Get bit value
                        tbit = irpBuffer[1];      

                        // Do a bit time slot
                        extension->CommResult[0] =          
                           docomm((UCHAR) (tbit ? 0xFF : 0xFE), 
                                           extension->Controller[PN]); 
                     }
                     else
                        extension->CommResult[0] = 1;     
                  break;

                  case DO_BYTE:          
                     // Get byte value
                     tbyte = irpBuffer[1];       
 
                     // See if we are on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         
                        // Initialize receive byte
                        rbyte = 0;                  
   
                        for (i = 0; i < 8; i++)      
                        {
                           rbyte >>= 1;               
   
                           // Read a bit and or it in
                           rbyte |= (docomm((UCHAR) (tbyte & 1 ? 0xFF : 0xFE), 
                                             extension->Controller[PN]) << 7); 
   
                           // Position transmit byte
                           tbyte >>= 1;              
                        }
                     }
                     else
                        rbyte = tbyte; // If bad port return byte received

                     // Store received byte
                     extension->CommResult[0] = rbyte;     
                  break;

                  case DO_BLOCK:
                     // See if we are on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         

                        // Save byte count
                        extension->CRLen = irpBuffer[1];
                        for (j = 0; j < extension->CRLen; j++)
                        {
                           extension->CommResult[j] = 
                                        FastByte(irpBuffer[j+2], 
                                                 extension->Controller[PN]);
                        }
                     }
                     else
                     {
                        for (i = 0; i < irpBuffer[1]; i++)
                           extension->CommResult[i] = irpBuffer[i+2];
                     }
                  break;

                  case FAST_ACCESS:
                     extension->CommResult[0] = FALSE;

                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         
   
                        if (!docomm(DOW_RESET, extension->Controller[PN]))
                        {
                           // Weak access succeeds if presence is detected
                           extension->CommResult[0] = TRUE;
                           FastByte(DOW_ACCESS, extension->Controller[PN]);

                           for (i = 0; i < 8; i++)      
                           {
                              FastByte(irpBuffer[i+1], 
                                       extension->Controller[PN]);
                           }
                        }
                     }
                  break;

                  case CHECK_OVERDRIVE:
                     // Are we on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         

                        // Send toggle overdrive signal 
                        extension->CommResult[0] =          
                           CheckOD(extension->Controller[PN]); 
                     }
                     else
                        extension->CommResult[0] = 0;     
                  break;

                  case TOGGLE_OVERDRIVE:
                     // Are we on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         

                        // Send toggle overdrive signal 
                        extension->CommResult[0] =          
                           dotoggle(extension->Controller[PN]); 
                     }
                     else
                        extension->CommResult[0] = 0;     
                  break;

                  case TOGGLE_PASSTHRU:
                     // Are we on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         

                        // Send 4 overdrive toggles
                        for (i = 0; i < 4; i++)
                           dotoggle(extension->Controller[PN]); 

                        extension->CommResult[0] = 1;
                     }
                     else
                        extension->CommResult[0] = 0;     
                  break;

                  case CHECK_BRICK:
                     // Are we on a valid port
                     if (extension->GoodPort)        
                     {
                        // Get current port number
                        PN = extension->pn;         

                        // Do a reset to invoke last part detect
                        docomm(DOW_RESET, extension->Controller[PN]);

                        // Send brick check sequence 
                        extension->CommResult[0] =          
                           docheck(extension->Controller[PN]); 
                     }
                     else
                        extension->CommResult[0] = 0;     
                  break;

                  case SET_PORT:                     
                     // Get port number
                     APtr = irpBuffer[1];       

                     // Is this a good # 
                     if (extension->ParNumFound >= APtr)   
                     { 
                        // Set new port number (0 based)
                        extension->pn = APtr - 1;       
                        // Indicate last port good 
                        extension->GoodPort = TRUE; 
                        
                        extension->CommResult[0] = 1;   
                     }
                     else
                     { 
                        // Indicate last port bad
                        extension->GoodPort = FALSE; 
                        // Fail if invalid port #
                        extension->CommResult[0] = 0;   
                     }
                  break;
               }
      
               OIrpSp->Parameters.Write.Length = 0; 
               OIrp->IoStatus.Information      = 2L;     
               OIrp->IoStatus.Status           = STATUS_SUCCESS; 
            }
            else
               status = STATUS_INVALID_PARAMETER; // Wrong number of bytes
         }

      break;

      default:
         status = STATUS_INVALID_PARAMETER; // Not an entry we support
   }

   // Wer'e done with this IRP
   IoCompleteRequest(OIrp, IO_NO_INCREMENT);        
   return status;
}

/////////////////////////////////////////////////////////////////////////////
//
//   Send toggle overdrive sequence to DS1481. 
//
UCHAR dotoggle(PUCHAR PortAddr)
{
   UCHAR bpap2;                            
   UCHAR BitVal;                          
   KIRQL oldIrql; // Current irterrupt request level 
   UCHAR ControlRegGlobal;

   // Keep spooler from running during a time slot
   KeRaiseIrql(CLOCK2_LEVEL, &oldIrql);        

   // 
   // Added in 2.54 to ensure ENI\ is high when D/CLK\ and RES\ are 
   // set to their initial state
   //
   WRITE_PORT_UCHAR(PortAddr, 0xDC);    
   KeStallExecutionProcessor(2);   

   // Set initial state of data lines for toggle overdrive
   WRITE_PORT_UCHAR(PortAddr, 0xFC);    

   KeStallExecutionProcessor(2);   

   // Save control register
   //bpap2 = (READ_PORT_UCHAR(PortAddr + 2) | 0x04) & 0x1C;        
   ControlRegGlobal = (READ_PORT_UCHAR(PortAddr + 2));
   bpap2 = (ControlRegGlobal | 0x04) & 0x1C;

   // Start time slot
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 | 0x02));  

   KeStallExecutionProcessor(4);                             

   BitVal = ((READ_PORT_UCHAR(PortAddr + 1) ^ 0x80) & 0x90) ? TRUE : FALSE;
   
   // Drive ENI\ back high
   //WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 & 0xFD));    

   // Drive a couple to our extra cs lines low 
   WRITE_PORT_UCHAR(PortAddr, NAND_MASK);    

   KeStallExecutionProcessor(2);              

   WRITE_PORT_UCHAR(PortAddr+2, (UCHAR) (ControlRegGlobal &CONTROL_REG_MASK));
   // let spooler back in
   KeLowerIrql(oldIrql);                                
   // Give 14 a little time to rise 
   KeStallExecutionProcessor(5);              

   return BitVal;                            
}

/////////////////////////////////////////////////////////////////////////////
//
//   Check to see if attached DS1481s are in overdrive. This function does
// not alter the state of the overdrive FF.
//
UCHAR CheckOD(PUCHAR PortAddr)
{
   UCHAR bpap2;                            
   KIRQL oldIrql; 
   int   i;
   UCHAR Busy      = FALSE;
   UCHAR OverDrive = FALSE;
   UCHAR ControlRegGlobal;
   // Keep spooler from running during a time slot
   KeRaiseIrql(CLOCK2_LEVEL, &oldIrql);        

   // 
   // Added in 2.54 to ensure ENI\ is high when D/CLK\ and RES\ are 
   // set to their initial state
   //
   WRITE_PORT_UCHAR(PortAddr, 0xDF);    
   KeStallExecutionProcessor(2);   

   // Set initial state of data lines
   WRITE_PORT_UCHAR(PortAddr, 0xFF);    

   // Save control register
   //bpap2 = (READ_PORT_UCHAR(PortAddr + 2) | 0x04) & 0x1C;        
   ControlRegGlobal = (READ_PORT_UCHAR(PortAddr +2));
   bpap2 = (ControlRegGlobal | 0x04) & 0x1C;

   // Start time slot
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 | 0x02));  

   // Wait for busy to pass for overdrive parts
   KeStallExecutionProcessor(16);                             

   OverDrive = ((READ_PORT_UCHAR(PortAddr+1)^0x80)&0x90) ? TRUE : FALSE;

   // let spooler back in
   KeLowerIrql(DISPATCH_LEVEL);                                

   i = 0;
   // Wait for time slot to finish
   while (!((READ_PORT_UCHAR(PortAddr+1)^0x80)&0x90) && (i++ < MAX_WAIT_TS))
      KeStallExecutionProcessor(4);  

   // Drive ENI\ back high
   //WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 & 0xFD));    

   // Drive a couple to our extra cs lines low 
   WRITE_PORT_UCHAR(PortAddr, NAND_MASK);
   KeStallExecutionProcessor(2);

   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR)(ControlRegGlobal & CONTROL_REG_MASK));

   // let spooler back in
   KeLowerIrql(oldIrql);                                
   // Give 14 a little time to rise 
   KeStallExecutionProcessor(5);              

   return OverDrive;                            
}

/////////////////////////////////////////////////////////////////////////////
//
//   Check for the presence of an adapter. 
//
UCHAR docheck(PUCHAR PortAddr)
{
   UCHAR bpap2;                            
   KIRQL oldIrql; // Current irterrupt request level 
   UCHAR Busy = FALSE;
   int   i;
   UCHAR ControlRegGlobal;

   // Keep spooler from running during a time slot
   KeRaiseIrql(DISPATCH_LEVEL, &oldIrql);        

   // 
   // Added in 2.54 to ensure ENI\ is high when D/CLK\ and RES\ are 
   // set to their initial state
   //
   WRITE_PORT_UCHAR(PortAddr, 0xDF);    
   KeStallExecutionProcessor(2);   

   // Set initial state of data lines
   WRITE_PORT_UCHAR(PortAddr, 0xFF);    

   KeStallExecutionProcessor(2);                             

   // Save control register
   //bpap2 = (READ_PORT_UCHAR(PortAddr + 2) | 0x04) & 0x1C;        
   ControlRegGlobal = (READ_PORT_UCHAR(PortAddr + 2));
   bpap2 = (ControlRegGlobal | 0x04) & 0x1C;

   // Start time slot
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 | 0x02));  

   // Wait for busy
   KeStallExecutionProcessor(4);                             

   Busy = ((READ_PORT_UCHAR(PortAddr + 1) ^ 0x80) & 0x90) ? FALSE : TRUE;

   if (Busy)
   {
      i = 0;                      
      while (!((READ_PORT_UCHAR(PortAddr+1) ^ 0x80) & 0x90) && (i++ < 25))
         KeStallExecutionProcessor(4);  

      if (i > 24)
         Busy = FALSE;
   }

   // Drive ENI\ back high
   //WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 & 0xFD));    

   // Drive a couple to our extra cs lines low 
   WRITE_PORT_UCHAR(PortAddr, NAND_MASK);    
   KeStallExecutionProcessor(2);
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR)(ControlRegGlobal & CONTROL_REG_MASK));

   // let spooler back in
   KeLowerIrql(oldIrql);                                
   // Give 14 a little time to rise 
   KeStallExecutionProcessor(5);              

   return Busy;                            
}

/////////////////////////////////////////////////////////////////////////////
//
//   This function manages the communication with the DS1410E(s).
//
UCHAR docomm(UCHAR com_byte, PUCHAR PortAddr)
{
   UCHAR bpap2;                            
   UCHAR BitVal;                          
   KIRQL oldIrql; // Current irterrupt request level 
   int   i;
   UCHAR ControlRegGlobal;

   // Keep spooler from running during a time slot
   KeRaiseIrql(DISPATCH_LEVEL, &oldIrql);        

   // 
   // Added in 2.54 to ensure ENI\ is high when D/CLK\ and RES\ are 
   // set to their initial state
   //
   WRITE_PORT_UCHAR(PortAddr, com_byte & 0xDF);    
   KeStallExecutionProcessor(2);   
   
   // Set initial state of data lines
   WRITE_PORT_UCHAR(PortAddr, com_byte);    

   KeStallExecutionProcessor(2);   

   // Save control register
   //bpap2 = (READ_PORT_UCHAR(PortAddr + 2) | 0x04) & 0x1C;        
   ControlRegGlobal = (READ_PORT_UCHAR(PortAddr + 2));
   bpap2 = (ControlRegGlobal | 0x04) & 0x1C;

   // Start time slot
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 | 0x02));  

   // Wait for busy
   KeStallExecutionProcessor(2);                             

   // Drive CLK\ high 
   WRITE_PORT_UCHAR(PortAddr, 0xFF);                        

   i = 0;                      
   while (!((READ_PORT_UCHAR(PortAddr + 1) ^ 0x80) & 0x90) && (i++ < MAX_WAIT))
      KeStallExecutionProcessor(1);  

   // Drive clk line low to get our bit
   WRITE_PORT_UCHAR(PortAddr, 0xFE);      
   // Wait for valid data 
   KeStallExecutionProcessor(2);                        
   BitVal = ((READ_PORT_UCHAR(PortAddr + 1) ^ 0x80) & 0x90) ? TRUE : FALSE;
   
   // Drive ENI\ back high
   //WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 & 0xFD));    
   

   // Drive a couple to our extra cs lines low 
   WRITE_PORT_UCHAR(PortAddr, NAND_MASK);    
   KeStallExecutionProcessor(2);
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR)(ControlRegGlobal & CONTROL_REG_MASK));

   // let spooler back in
   KeLowerIrql(oldIrql);                                
   // Give 14 a little time to rise 
   KeStallExecutionProcessor(3);              

   return BitVal;                            
}

/////////////////////////////////////////////////////////////////////////////
//
//   This function sends an entire byte to the DS1481. It should only be
// called in overdrive mode !!! 
//
UCHAR FastByte(UCHAR SB, PUCHAR PortAddr)
{
   UCHAR bpap2;                            
   UCHAR BitVal, ByteVal;                          
   UCHAR com_byte;
   KIRQL oldIrql; // Current irterrupt request level 
   int   i, j;
   UCHAR ControlRegGlobal;

   // Keep spooler from running during a time slot
   KeRaiseIrql(DISPATCH_LEVEL, &oldIrql);        
   // Save control register
   //bpap2 = (READ_PORT_UCHAR(PortAddr + 2) | 0x04) & 0x1C;        
   ControlRegGlobal = (READ_PORT_UCHAR(PortAddr + 2));
   bpap2 = (ControlRegGlobal | 0x04) & 0x1C;

   ByteVal = 0;
   for (j = 0; j < 8; j++)
   {
      com_byte = (SB & 1) ? WRITE_1 : WRITE_0;
      SB >>= 1;

      // D0-7 = 0xFE, don't do slow out if unneccessary (for j > 0)
      if ((!j) || (com_byte == WRITE_1))
      {
         // Set initial state of data lines
         WRITE_PORT_UCHAR(PortAddr, com_byte);    
      }
   
      // Start time slot
      WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 | 0x02));  
   
      // Wait for busy
      KeStallExecutionProcessor(3);                             
   
      // Drive CLK\ high 
      WRITE_PORT_UCHAR(PortAddr, 0xFF);                        
   
      i = 0;                      
      while (!((READ_PORT_UCHAR(PortAddr+1)^0x80)&0x90) && (i++<MAX_WAIT))
         KeStallExecutionProcessor(1);  
   
      // Drive clk line low to get our bit
      WRITE_PORT_UCHAR(PortAddr, 0xFE);      
      BitVal = ((READ_PORT_UCHAR(PortAddr + 1) ^ 0x80) & 0x90) ? TRUE : FALSE;
      
      // Drive ENI\ back high
      WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR) (bpap2 & 0xFD));    
      KeStallExecutionProcessor(1);

      ByteVal >>= 1;
      if (BitVal)
         ByteVal |= 0x80;
   }

   // All data lines should be left high
//   WRITE_PORT_UCHAR(PortAddr, 0xFF);
   WRITE_PORT_UCHAR(PortAddr, NAND_MASK);
   KeStallExecutionProcessor(2);
   WRITE_PORT_UCHAR(PortAddr + 2, (UCHAR)(ControlRegGlobal & CONTROL_REG_MASK));

   // Let spooler back in
   KeLowerIrql(oldIrql);

   return ByteVal;
}

